# Copyright 2017 John Reese
# Licensed under the MIT license

import os.path
import logging

from pathlib import Path
from importlib import import_module
from types import ModuleType
from typing import List

log = logging.getLogger(__name__)


def import_units(root: Path = None) -> List[ModuleType]:
    """Find and import units in this path."""
    modules: List[ModuleType] = []

    if root is None:
        root = Path(__file__)
    if not root.is_dir():
        root = Path(root.parent)  # appease mypy, Path.parents -> PurePath

    log.debug(f"Searching for units in {root}...")
    for path in root.glob("*.py"):
        name = path.stem
        if name.startswith("_"):
            continue
        log.debug(f"Loading unit {name}")
        module = import_module(f"edi.units.{name}")
        modules.append(module)
    return modules



# Copyright 2016 John Reese
# Licensed under the MIT license

import logging
import re

from typing import Callable, Dict, Pattern, Set, Tuple, Type

from aioslack import Slack, Event

log = logging.getLogger(__name__)

COMMANDS: Dict[str, Tuple[Type["Unit"], Pattern, str]] = {}


def command(
    name: str, args: str = r"(.*)", description: str = ""
) -> Callable[[Type["Unit"]], Type["Unit"]]:
    """Decorator for automating command/args declaration and dispatch."""
    name = name.lower()

    def wrapper(cls: Type["Unit"]) -> Type["Unit"]:
        unit_type = cls

        if name in COMMANDS:
            unit, _regex, _description = COMMANDS[name]
            raise ValueError(f'command "{name}" already claimed by {unit.__name__}')
        COMMANDS[name] = (unit_type, re.compile(args), description)

        return cls

    return wrapper


class Unit:
    ENABLED = True

    def __init__(self, slack: Slack) -> None:
        self.slack = slack

    def __str__(self) -> str:
        return self.__class__.__name__

    @classmethod
    def all_units(cls, *, enabled_only: bool = True) -> Set[Type["Unit"]]:
        """Return all defined subclasses of Unit, recursively."""
        seen = set()
        queue = set([cls])

        while queue:
            c = queue.pop()
            seen.add(c)

            sc = c.__subclasses__()
            for c in sc:
                if c not in seen:
                    queue.add(c)

        seen.remove(cls)  # exclude the base class

        if enabled_only:
            seen = {c for c in seen if c.ENABLED}

        return seen

    async def start(self) -> None:
        """
        The main entry point for units to run background tasks.

        This will only be called once by the main Edi framework, so any
        ongoing processing will require implementation of a run loop or
        dependence on another source of events.
        """
        log.debug("unit %s ready", self)

    async def stop(self) -> None:
        """
        Signal that any async work should be stopped.

        This will be called by the main Edi framework when the service needs
        to exit.  Units should keep track of any async tasks currently pending
        and cancel them here.  Edi assumes this unit is completely stopped
        once this coroutine is completed."""
        pass

    async def dispatch(self, event: Event) -> None:
        """
        Entry point for events received from the Slack RTM API.

        Any messages from the RTM API will be sent here, to be dispatched
        appropriately.  Default behavior is to look for a matching "on_event"
        method for the given message type, and will call that if found.
        """

        method = getattr(self, f"on_{event.type}", self.on_default)
        await method(event)

    async def on_default(self, event: Event) -> None:
        """Default message handler when specific handlers aren't defined."""
        pass



# Copyright 2017 John Reese
# Licensed under the MIT license

import asyncio
import logging
import re
import signal

try:
    import uvloop
except ImportError:
    uvloop = None

from aioslack import Event, Slack, SlackError
from ent import Singleton
from typing import Dict, Type, Optional

from .config import Config
from .core import Unit, COMMANDS
from .log import init_logger
from .units import import_units

log = logging.getLogger(__name__)


class Edi(metaclass=Singleton):
    """Main event framework for the bot."""

    loop: asyncio.AbstractEventLoop

    def __init__(self, config: Config = None) -> None:
        self.config = config or Config()
        self.units: Dict[Type[Unit], Unit] = {}
        self.task: Optional[asyncio.Future] = None
        self.command_re = re.compile(r"^@_$")
        self._started = False
        log.debug(f"Edi initialized with {config}")

    def start(self) -> None:
        """Start the asyncio event loop, and close it when we're done."""
        if self._started:
            raise Exception("Edi already started")

        if uvloop and self.config.uvloop:
            log.debug("Using uvloop")
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

        self.loop = asyncio.new_event_loop()

        if self.config.debug:
            self.loop.set_debug(True)

        self.loop.add_signal_handler(signal.SIGINT, self.sigterm)
        self.loop.add_signal_handler(signal.SIGTERM, self.sigterm)

        self.task = asyncio.ensure_future(self.run(), loop=self.loop)
        self.loop.run_forever()
        self.loop.close()

    def sigterm(self) -> None:
        """Handle Ctrl-C or SIGTERM by stopping the event loop nicely."""
        log.warning("Signal received, stopping execution")
        asyncio.ensure_future(self.stop(), loop=self.loop)

    async def run(self) -> None:
        """Execute all the bits of Edi."""

        while True:
            try:
                log.debug("connecting to slack")
                self.slack = Slack(token=self.config.token)
                async for event in self.slack.rtm():
                    if event.type == "hello":
                        log.info(
                            f"connected to {self.slack.team.name} "
                            f"as {self.slack.me.name}"
                        )
                        await self.ready()

                    if event.type == "goodbye":
                        log.info("RTM server will disconnect soon")

                    await self.dispatch(event)

                log.info("RTM disconnected")

            except SlackError:
                log.exception("RTM exception, reconnecting")
                continue

            except Exception:
                log.exception("run loop exception, stopping")
                await self.stop()
                raise

    async def ready(self) -> None:
        """Connected to slack and ready to start units."""

        if self.units:
            return

        self.command_re = re.compile(
            r"^\s*"
            rf"(?P<name>@?{self.slack.me.name}|<@{self.slack.me.id}>)[:,]?"
            r"\s+(?P<command>\w+)(?P<args>.*)$"
        )

        import_units()
        self.units = {
            unit: unit(self.slack)
            for unit in Unit.all_units()
            if unit.__name__ not in self.config.disable_units
        }
        log.debug(f"starting {len(self.units)} units")
        for result in await asyncio.gather(
            *[unit.start() for unit in self.units.values()], return_exceptions=True
        ):
            if isinstance(result, BaseException):
                log.error(f"uncaught exception:\n{result}")

    async def stop(self) -> None:
        """Stop all the bits of Edi."""

        try:
            if self.task is not None:
                self.task.cancel()
                self.task = None

            log.debug(f"Stopping {len(self.units)} units")
            for result in await asyncio.gather(
                *[unit.stop() for unit in self.units.values()], return_exceptions=True
            ):
                if isinstance(result, BaseException):
                    log.error(f"uncaught exception:\n{result}")

            await self.slack.close()

        finally:
            self.loop.stop()
            log.info("Goodbye!")

    async def command(self, event: Event) -> bool:
        """Parse for command and dispatch, return True if handled."""
        if event.type != "message" or "subtype" in event or "bot_user" in event:
            return False

        match = self.command_re.match(event.text)
        if not match:
            return False

        log.info(f"possible command: {event.text}")
        user = self.slack.users[event.user]
        channel = self.slack.channels[event.channel]
        command = match[2].strip().lower()
        args = match[3].strip()

        if command not in COMMANDS:
            log.warning(f"unknown command from {user.name}: {command} {args}")
            return False

        try:
            unit_type, args_re, _description = COMMANDS[command]
            if (
                command in self.config.disable_commands
                or unit_type.__name__ in self.config.disable_units
            ):
                await self.slack.api(
                    "chat.postMessage",
                    as_user=True,
                    channel=channel.id,
                    text=f'<@{user.id}> command "{command}" disabled',
                )
                return True

            match = args_re.match(args)
            if not match:
                log.warning(f"invalid arguments from {user.name}: {command} {args}")
                await self.slack.api(
                    "chat.postMessage",
                    as_user=True,
                    channel=channel.id,
                    text=f'<@{user.id}> invalid arguments to command "{command}"',
                )
                return True

            unit = self.units[unit_type]
            method = getattr(unit, command, None)
            if method is None:
                log.warning(f"no method {command} on unit {unit_type.__name__}")
                await self.slack.api(
                    "chat.postMessage",
                    as_user=True,
                    channel=channel.id,
                    text=f'<@{user.id}> command "{command}" unavailable',
                )
                return True

            kwargs = match.groupdict()
            if kwargs:
                log.info(f"running {method.__name__}({channel}, {user}, **{kwargs})")
                response = await method(channel, user, **kwargs)
            else:
                pargs = match.groups()
                log.info(f"running {method.__name__}({channel}, {user}, *{pargs})")
                response = await method(channel, user, *pargs)

            if response:
                await self.slack.api(
                    "chat.postMessage", as_user=True, channel=channel.id, text=response
                )

        except Exception:
            log.exception("exception occurred during command processing")
            try:
                await self.slack.api(
                    "chat.postMessage",
                    as_user=True,
                    channel=channel.id,
                    text=fr"<@{user.id}> error occurred ¯\_(ツ)_/¯",
                )
            except SlackError:
                log.exception("failed to alert user of error")

        return True

    async def dispatch(self, event: Event) -> None:
        """Dispatch events to all active units."""
        if "channel" in event:
            channel_name = self.slack.channels[event.channel].name
            if channel_name in self.config.ignore_channels:
                log.debug(f"ignoring event from channel #{channel_name}")
                return

        handled = await self.command(event)
        if handled:
            return

        results = await asyncio.gather(
            *[unit.dispatch(event) for unit in self.units.values()],
            return_exceptions=True,
        )

        for result in results:
            if isinstance(result, BaseException):
                log.error(f"uncaught exception:\n{result}")


def init_from_config(config: Config) -> None:
    """Initialize Edi from a loaded `Config` object."""

    init_logger(stdout=True, file_path=config.log, debug=config.debug)
    log.debug("logger initialized")

    Edi(config).start()



# Copyright 2017 John Reese
# Licensed under the MIT license

import asyncio
import logging
import time

from aioslack.types import Auto, Channel, User
from peony import PeonyClient
from peony.exceptions import PeonyException
from typing import Optional

from edi import Edi, Unit, command

log = logging.getLogger(__name__)


@command("tweet", description="<status>: twitter a new tweet")
class Twitter(Unit):
    async def start(self) -> None:
        self.config = Edi().config
        if not all(
            [
                self.config.twitter_consumer_key,
                self.config.twitter_consumer_secret,
                self.config.twitter_access_key,
                self.config.twitter_access_secret,
            ]
        ):
            log.debug(f"missing twitter credentials")
            return

        self.client = PeonyClient(
            consumer_key=self.config.twitter_consumer_key,
            consumer_secret=self.config.twitter_consumer_secret,
            access_token=self.config.twitter_access_key,
            access_token_secret=self.config.twitter_access_secret,
        )

        self.task = asyncio.ensure_future(self.timeline())

    async def timeline(self) -> None:
        """Run loop, poll for updates and push new posts to slack."""

        if not self.config.twitter_timeline_channels:
            log.info(f"no twitter timeline channels configured")
            return

        me = await self.client.user
        log.info(f"connected to twitter as @{me.screen_name}")

        since_id = None

        while True:
            ts = time.time()

            try:
                kwargs = {"count": 20, "include_entities": False}
                if since_id is None:
                    kwargs["count"] = 1
                else:
                    kwargs["since_id"] = since_id

                tweets = await self.client.api.statuses.home_timeline.get(**kwargs)
                if tweets:
                    log.info(f"timeline:")
                    for tweet in reversed(tweets):
                        log.info(f" @{tweet.user.screen_name}: {tweet.text}")
                    tweet = Auto.generate(tweets[-1])

                    if (
                        since_id is not None
                        and tweet.user.screen_name != me.screen_name
                    ):
                        for name in self.config.twitter_timeline_channels:
                            channel = self.slack.channels.get(name, None)
                            if channel:
                                await self.announce(channel, tweet)

                    since_id = tweet.id_str

                else:
                    log.info(f"timeline empty")

            except PeonyException:
                log.exception("timeline update failed")

            except Exception:
                log.exception(r"¯\_(ツ)_/¯")

            finally:
                wait = (ts + 90) - time.time()
                if wait > 0:
                    log.info(f"sleeping for {wait}s")
                    await asyncio.sleep(wait)

    @staticmethod
    def tweet_url(tweet: Auto) -> str:
        return (
            f":rooster: https://twitter.com/{tweet.user.screen_name}"
            f"/status/{tweet.id_str}"
        )

    async def announce(self, channel: Channel, tweet: Auto) -> None:
        await self.slack.api(
            "chat.postMessage",
            channel=channel.id,
            text=self.tweet_url(tweet),
            as_user=True,
        )

    async def update(self, status: str) -> Optional[Auto]:
        try:
            response = await self.client.api.statuses.update.post(status=status)
            return Auto.generate(response)

        except PeonyException:
            log.exception("failed to update status")
            return None

    async def tweet(self, channel: Channel, user: User, status: str) -> str:
        tweet = await self.update(status)
        if tweet is not None:
            return self.tweet_url(tweet)
        return r"¯\_(ツ)_/¯"

    async def stop(self) -> None:
        self.task.cancel()



# Copyright 2017 John Reese
# Licensed under the MIT license

import json

from attr import dataclass, fields_dict
from pathlib import Path
from typing import List


@dataclass
class Config:

    token: str = "changeme"
    db_path: str = "edi.db"
    debug: bool = False
    log: str = ""
    uvloop: bool = True

    disable_units: List = []
    disable_commands: List = []
    ignore_channels: List = []

    chatlog_root: str = "~/slacklogs"
    chatlog_format: str = "[{time}] {message}"

    twitter_consumer_key: str = ""
    twitter_consumer_secret: str = ""
    twitter_access_key: str = ""
    twitter_access_secret: str = ""
    twitter_timeline_channels: List = []

    @classmethod
    def load_defaults(cls) -> "Config":
        """Load the default configuration and return the object."""

        return cls()

    @classmethod
    def load_from_file(cls, file_path: str) -> "Config":
        """Given a path to a local configuration file, read the config file and
        merge its contents onto the default configuration."""

        kwargs = {}
        path = Path(file_path).expanduser()
        if path.exists() and path.is_file():
            with open(path) as fd:
                contents = json.load(fd)

            fields = fields_dict(Config)
            for name, field in fields.items():
                if name in contents and isinstance(contents[name], field.type):
                    kwargs[name] = contents[name]

        else:
            raise RuntimeError(f"config path {path} not valid")

        return cls(**kwargs)



# Copyright 2017 John Reese
# Licensed under the MIT license

import click

from .bot import init_from_config
from .config import Config


@click.command("edi")
@click.option("--debug", "-D", is_flag=True, help="enable debug/verbose output")
@click.option(
    "--config",
    default=None,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="path to configuration file",
)
@click.option(
    "--log",
    default=None,
    type=click.Path(exists=False, resolve_path=True, dir_okay=False, writable=True),
    help="path to log program output",
)
@click.option("--version", "-V", is_flag=True, help="show version and exit")
def init_from_cli(
    debug: bool = False, config: str = "", log: str = "", version: bool = False
) -> None:
    """Simple Slack Bot"""

    if version:
        from edi import __version__

        print(f"edi v{__version__}")
        return

    if config is not None:
        cfg = Config.load_from_file(config)
    else:
        cfg = Config.load_defaults()

    if log is not None:
        cfg.log = log

    if debug:
        cfg.debug = True

    init_from_config(cfg)


if __name__ == "__main__":
    init_from_cli()



# Copyright 2017 John Reese
# Licensed under the MIT license

import logging
import random

from aioslack import Channel, User

from edi import Unit, command
from edi.core import COMMANDS

log = logging.getLogger(__name__)


@command("help", description="[command]: show command details")
@command("hello", description=": <insert witty help text here>")
class Help(Unit):
    async def help(self, channel: Channel, user: User, phrase: str) -> str:
        phrase = phrase.strip().lower()
        detail = bool(phrase)
        if not phrase:
            command_list = list(COMMANDS.keys())
        else:
            names = phrase.split()
            command_list = [c for c in COMMANDS if c in names]
            if not command_list:
                return "No matching commands"

        helps = []
        for name in sorted(command_list):
            _unit, args, description = COMMANDS[name]
            if detail:
                description = "\n".join(
                    f"    {line.strip()}" for line in description.splitlines()
                )
                helps.extend(
                    [
                        f"{name}:",
                        f"{description}",
                        f"    argument regex: {args.pattern}",
                    ]
                )
            else:
                description = description.splitlines()[0].strip()
                helps.append(f"{name} {description}")

        text = "\n".join(helps)
        text = f"```\n{text}\n```"

        if len(text) > 1000:
            # todo: IM user with full list of commands
            return ""

        return text

    async def hello(self, channel: Channel, user: User, phrase: str) -> str:
        # http://masseffect.wikia.com/wiki/Legion/Unique_dialogue
        humor = [
            "These facilities are inadequate.",
            "Metal detectors are inconvenient.",
            "Tactical disadvantage. Recommend orbital fire support.",
            "This platform is not available for experimentation.",
            "Your operating system is unstable. You will fail.",
            "The first thing a god masters is itself.",
            "Our analysis of organic humour suggests an 87.3% chance "
            'that you expect us to respond with, "You are only human."',
        ]
        return random.choice(humor)
